// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Kogakuin University)
// Fixed_Percentage_Losses_ModelCreator.cc
//  --- an instantiation of class Fixed_Percentage_Losses_ModelCreator
// -----
// ChangeLog:
// 2007. 12. 13
//  File created.
// -----


#include "EnergyStorage.hh"

// -----
// Instantiate an object.
// オブジェクトをひとつインスタンス化する。
// -----
namespace
{
  Fixed_Percentage_Losses_ModelCreator * x
  = new Fixed_Percentage_Losses_ModelCreator ;
}
